/*
===========================================================================
Copyright (C) 2007-2011 Robert Beckebans <trebor_7@users.sourceforge.net>

This file is part of Daemon source code.

Daemon source code is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of the License,
or (at your option) any later version.

Daemon source code is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Daemon source code; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
===========================================================================
*/
// tr_vbo.c
#include "tr_local.h"

// "templates" for VBO vertex data layouts

// interleaved position and qtangents per frame/vertex in part 1
struct fmtVertexAnim1 {
	i16vec4_t position;
	i16vec4_t qtangents;
};
const GLsizei sizeVertexAnim1 = sizeof( struct fmtVertexAnim1 );
// interleaved texcoords and colour in part 2
struct fmtVertexAnim2 {
	i16vec2_t texcoord;
	u8vec4_t  colour;
};
const GLsizei sizeVertexAnim2 = sizeof( struct fmtVertexAnim2 );

// interleaved data: position, texcoord, colour, qtangent, bonefactors
struct fmtSkeletal {
	i16vec4_t position;
	i16vec2_t texcoord;
	u8vec4_t  colour;
	i16vec4_t qtangents;
	u16vec4_t boneFactors;
};
const GLsizei sizeSkeletal = sizeof( struct fmtSkeletal );

// interleaved data: position, colour, qtangent, texcoord
// -> struct shaderVertex_t in tr_local.h
const GLsizei sizeShaderVertex = sizeof( shaderVertex_t );


static uint32_t R_DeriveAttrBits( vboData_t data )
{
	uint32_t stateBits = 0;

	if ( data.xyz )
	{
		stateBits |= ATTR_POSITION;
	}

	if ( data.qtangent )
	{
		stateBits |= ATTR_QTANGENT;
	}

	if ( data.color )
	{
		stateBits |= ATTR_COLOR;
	}

	if ( data.st )
	{
		stateBits |= ATTR_TEXCOORD;
	}

	if ( data.boneIndexes && data.boneWeights )
	{
		stateBits |= ATTR_BONE_FACTORS;
	}

	if ( data.numFrames )
	{
		if ( data.xyz )
		{
			stateBits |= ATTR_POSITION2;
		}

		if ( data.qtangent )
		{
			stateBits |= ATTR_QTANGENT2;
		}
	}

	return stateBits;
}

static void R_SetVBOAttributeComponentType( VBO_t *vbo, uint32_t i, qboolean noLightCoords )
{
	if ( i == ATTR_INDEX_BONE_FACTORS )
	{
		vbo->attribs[ i ].componentType = GL_UNSIGNED_SHORT;
	}
	else if ( i == ATTR_INDEX_COLOR )
	{
		vbo->attribs[ i ].componentType = GL_UNSIGNED_BYTE;
	}
	else if ( i == ATTR_INDEX_TEXCOORD )
	{
		vbo->attribs[ i ].componentType = GL_HALF_FLOAT;
	}
	else if ( i == ATTR_INDEX_QTANGENT || i == ATTR_INDEX_QTANGENT2 )
	{
		vbo->attribs[ i ].componentType = GL_SHORT;
	}
	else
	{
		vbo->attribs[ i ].componentType = GL_FLOAT;
	}

	if ( i == ATTR_INDEX_COLOR || i == ATTR_INDEX_QTANGENT || i == ATTR_INDEX_QTANGENT2 )
	{
		vbo->attribs[ i ].normalize = GL_TRUE;
	}
	else
	{
		vbo->attribs[ i ].normalize = GL_FALSE;
	}
	
	if ( i == ATTR_INDEX_TEXCOORD && noLightCoords )
	{
		vbo->attribs[ i ].numComponents = 2;
	}
	else if ( i == ATTR_INDEX_POSITION || i == ATTR_INDEX_POSITION2 )
	{
		vbo->attribs[ i ].numComponents = 3;
	}
	else
	{
		vbo->attribs[ i ].numComponents = 4;
	}
}

static void R_SetAttributeLayoutsVertexAnimation( VBO_t *vbo )
{
	// part 1 is repeated for every frame
	GLsizei sizePart1 = sizeVertexAnim1 * vbo->vertexesNum * vbo->framesNum;
	GLsizei sizePart2 = sizeVertexAnim2 * vbo->vertexesNum;

	vbo->attribs[ ATTR_INDEX_POSITION ].numComponents = 4;
	vbo->attribs[ ATTR_INDEX_POSITION ].componentType = GL_SHORT;
	vbo->attribs[ ATTR_INDEX_POSITION ].normalize     = GL_TRUE;
	vbo->attribs[ ATTR_INDEX_POSITION ].ofs           = offsetof( struct fmtVertexAnim1, position );
	vbo->attribs[ ATTR_INDEX_POSITION ].realStride    = sizeVertexAnim1;
	vbo->attribs[ ATTR_INDEX_POSITION ].stride        = sizeVertexAnim1;
	vbo->attribs[ ATTR_INDEX_POSITION ].frameOffset   = sizeVertexAnim1 * vbo->vertexesNum;

	vbo->attribs[ ATTR_INDEX_QTANGENT ].numComponents = 4;
	vbo->attribs[ ATTR_INDEX_QTANGENT ].componentType = GL_SHORT;
	vbo->attribs[ ATTR_INDEX_QTANGENT ].normalize     = GL_TRUE;
	vbo->attribs[ ATTR_INDEX_QTANGENT ].ofs          = offsetof( struct fmtVertexAnim1, qtangents );
	vbo->attribs[ ATTR_INDEX_QTANGENT ].realStride   = sizeVertexAnim1;
	vbo->attribs[ ATTR_INDEX_QTANGENT ].stride       = sizeVertexAnim1;
	vbo->attribs[ ATTR_INDEX_QTANGENT ].frameOffset  = sizeVertexAnim1 * vbo->vertexesNum;

	// these use the same layout
	vbo->attribs[ ATTR_INDEX_POSITION2 ] = vbo->attribs[ ATTR_INDEX_POSITION ];
	vbo->attribs[ ATTR_INDEX_QTANGENT2 ] = vbo->attribs[ ATTR_INDEX_QTANGENT ];

	vbo->attribs[ ATTR_INDEX_TEXCOORD ].numComponents = 2;
	vbo->attribs[ ATTR_INDEX_TEXCOORD ].componentType = GL_HALF_FLOAT;
	vbo->attribs[ ATTR_INDEX_TEXCOORD ].normalize     = GL_FALSE;
	vbo->attribs[ ATTR_INDEX_TEXCOORD ].ofs          = sizePart1 + offsetof( struct fmtVertexAnim2, texcoord );
	vbo->attribs[ ATTR_INDEX_TEXCOORD ].realStride   = sizeVertexAnim2;
	vbo->attribs[ ATTR_INDEX_TEXCOORD ].stride       = sizeVertexAnim2;
	vbo->attribs[ ATTR_INDEX_TEXCOORD ].frameOffset  = 0;

	vbo->attribs[ ATTR_INDEX_COLOR ].numComponents   = 4;
	vbo->attribs[ ATTR_INDEX_COLOR ].componentType   = GL_UNSIGNED_BYTE;
	vbo->attribs[ ATTR_INDEX_COLOR ].normalize       = GL_TRUE;
	vbo->attribs[ ATTR_INDEX_COLOR ].ofs             = sizePart1 + offsetof( struct fmtVertexAnim2, colour );
	vbo->attribs[ ATTR_INDEX_COLOR ].realStride      = sizeVertexAnim2;
	vbo->attribs[ ATTR_INDEX_COLOR ].stride          = sizeVertexAnim2;
	vbo->attribs[ ATTR_INDEX_COLOR ].frameOffset     = 0;

	// total size
	vbo->vertexesSize = sizePart1 + sizePart2;
}

static void R_SetAttributeLayoutsSkeletal( VBO_t *vbo )
{
	vbo->attribs[ ATTR_INDEX_POSITION ].numComponents = 4;
	vbo->attribs[ ATTR_INDEX_POSITION ].componentType = GL_SHORT;
	vbo->attribs[ ATTR_INDEX_POSITION ].normalize     = GL_TRUE;
	vbo->attribs[ ATTR_INDEX_POSITION ].ofs           = offsetof( struct fmtSkeletal, position );
	vbo->attribs[ ATTR_INDEX_POSITION ].realStride    = sizeSkeletal;
	vbo->attribs[ ATTR_INDEX_POSITION ].stride        = sizeSkeletal;
	vbo->attribs[ ATTR_INDEX_POSITION ].frameOffset   = 0;

	vbo->attribs[ ATTR_INDEX_TEXCOORD ].numComponents = 2;
	vbo->attribs[ ATTR_INDEX_TEXCOORD ].componentType = GL_HALF_FLOAT;
	vbo->attribs[ ATTR_INDEX_TEXCOORD ].normalize     = GL_FALSE;
	vbo->attribs[ ATTR_INDEX_TEXCOORD ].ofs          = offsetof( struct fmtSkeletal, texcoord );
	vbo->attribs[ ATTR_INDEX_TEXCOORD ].realStride   = sizeSkeletal;
	vbo->attribs[ ATTR_INDEX_TEXCOORD ].stride       = sizeSkeletal;
	vbo->attribs[ ATTR_INDEX_TEXCOORD ].frameOffset  = 0;

	vbo->attribs[ ATTR_INDEX_COLOR ].numComponents   = 4;
	vbo->attribs[ ATTR_INDEX_COLOR ].componentType   = GL_UNSIGNED_BYTE;
	vbo->attribs[ ATTR_INDEX_COLOR ].normalize       = GL_TRUE;
	vbo->attribs[ ATTR_INDEX_COLOR ].ofs             = offsetof( struct fmtSkeletal, colour );
	vbo->attribs[ ATTR_INDEX_COLOR ].realStride      = sizeSkeletal;
	vbo->attribs[ ATTR_INDEX_COLOR ].stride          = sizeSkeletal;
	vbo->attribs[ ATTR_INDEX_COLOR ].frameOffset     = 0;

	vbo->attribs[ ATTR_INDEX_QTANGENT ].numComponents = 4;
	vbo->attribs[ ATTR_INDEX_QTANGENT ].componentType = GL_SHORT;
	vbo->attribs[ ATTR_INDEX_QTANGENT ].normalize     = GL_TRUE;
	vbo->attribs[ ATTR_INDEX_QTANGENT ].ofs          = offsetof( struct fmtSkeletal, qtangents );
	vbo->attribs[ ATTR_INDEX_QTANGENT ].realStride   = sizeSkeletal;
	vbo->attribs[ ATTR_INDEX_QTANGENT ].stride       = sizeSkeletal;
	vbo->attribs[ ATTR_INDEX_QTANGENT ].frameOffset  = 0;

	vbo->attribs[ ATTR_INDEX_BONE_FACTORS ].numComponents = 4;
	vbo->attribs[ ATTR_INDEX_BONE_FACTORS ].componentType = GL_UNSIGNED_SHORT;
	vbo->attribs[ ATTR_INDEX_BONE_FACTORS ].normalize     = GL_FALSE;
	vbo->attribs[ ATTR_INDEX_BONE_FACTORS ].ofs           = offsetof( struct fmtSkeletal, boneFactors );
	vbo->attribs[ ATTR_INDEX_BONE_FACTORS ].realStride    = sizeSkeletal;
	vbo->attribs[ ATTR_INDEX_BONE_FACTORS ].stride        = sizeSkeletal;
	vbo->attribs[ ATTR_INDEX_BONE_FACTORS ].frameOffset   = 0;

	// total size
	vbo->vertexesSize = sizeSkeletal * vbo->vertexesNum;
}

static void R_SetAttributeLayoutsStatic( VBO_t *vbo )
{
	vbo->attribs[ ATTR_INDEX_POSITION ].numComponents = 3;
	vbo->attribs[ ATTR_INDEX_POSITION ].componentType = GL_FLOAT;
	vbo->attribs[ ATTR_INDEX_POSITION ].normalize     = GL_FALSE;
	vbo->attribs[ ATTR_INDEX_POSITION ].ofs           = offsetof( shaderVertex_t, xyz );
	vbo->attribs[ ATTR_INDEX_POSITION ].realStride    = sizeShaderVertex;
	vbo->attribs[ ATTR_INDEX_POSITION ].stride        = sizeShaderVertex;
	vbo->attribs[ ATTR_INDEX_POSITION ].frameOffset   = 0;

	vbo->attribs[ ATTR_INDEX_COLOR ].numComponents   = 4;
	vbo->attribs[ ATTR_INDEX_COLOR ].componentType   = GL_UNSIGNED_BYTE;
	vbo->attribs[ ATTR_INDEX_COLOR ].normalize       = GL_TRUE;
	vbo->attribs[ ATTR_INDEX_COLOR ].ofs             = offsetof( shaderVertex_t, color );
	vbo->attribs[ ATTR_INDEX_COLOR ].realStride      = sizeShaderVertex;
	vbo->attribs[ ATTR_INDEX_COLOR ].stride          = sizeShaderVertex;
	vbo->attribs[ ATTR_INDEX_COLOR ].frameOffset     = 0;

	vbo->attribs[ ATTR_INDEX_QTANGENT ].numComponents = 4;
	vbo->attribs[ ATTR_INDEX_QTANGENT ].componentType = GL_SHORT;
	vbo->attribs[ ATTR_INDEX_QTANGENT ].normalize     = GL_TRUE;
	vbo->attribs[ ATTR_INDEX_QTANGENT ].ofs           = offsetof( shaderVertex_t, qtangents );
	vbo->attribs[ ATTR_INDEX_QTANGENT ].realStride    = sizeShaderVertex;
	vbo->attribs[ ATTR_INDEX_QTANGENT ].stride        = sizeShaderVertex;
	vbo->attribs[ ATTR_INDEX_QTANGENT ].frameOffset   = 0;

	vbo->attribs[ ATTR_INDEX_TEXCOORD ].numComponents = 4;
	vbo->attribs[ ATTR_INDEX_TEXCOORD ].componentType = GL_HALF_FLOAT;
	vbo->attribs[ ATTR_INDEX_TEXCOORD ].normalize     = GL_FALSE;
	vbo->attribs[ ATTR_INDEX_TEXCOORD ].ofs           = offsetof( shaderVertex_t, texCoords );
	vbo->attribs[ ATTR_INDEX_TEXCOORD ].realStride    = sizeShaderVertex;
	vbo->attribs[ ATTR_INDEX_TEXCOORD ].stride        = sizeShaderVertex;
	vbo->attribs[ ATTR_INDEX_TEXCOORD ].frameOffset   = 0;

	// total size
	vbo->vertexesSize = sizeShaderVertex * vbo->vertexesNum;
}

static void R_SetAttributeLayoutsPosition( VBO_t *vbo )
{
	vbo->attribs[ ATTR_INDEX_POSITION ].numComponents = 3;
	vbo->attribs[ ATTR_INDEX_POSITION ].componentType = GL_FLOAT;
	vbo->attribs[ ATTR_INDEX_POSITION ].normalize     = GL_FALSE;
	vbo->attribs[ ATTR_INDEX_POSITION ].ofs           = 0;
	vbo->attribs[ ATTR_INDEX_POSITION ].realStride    = sizeof( vec3_t );
	vbo->attribs[ ATTR_INDEX_POSITION ].stride        = sizeof( vec3_t );
	vbo->attribs[ ATTR_INDEX_POSITION ].frameOffset   = 0;

	// total size
	vbo->vertexesSize = sizeof( vec3_t ) * vbo->vertexesNum;
}

static void R_SetVBOAttributeLayouts( VBO_t *vbo, qboolean noLightCoords )
{
	uint32_t i;
	for ( i = 0; i < ATTR_INDEX_MAX; i++ )
	{
		R_SetVBOAttributeComponentType( vbo, i, noLightCoords );
	}

	if ( vbo->layout == VBO_LAYOUT_VERTEX_ANIMATION )
	{
		R_SetAttributeLayoutsVertexAnimation( vbo );
	}
	else if ( vbo->layout == VBO_LAYOUT_SKELETAL )
	{
		R_SetAttributeLayoutsSkeletal( vbo );
	}
	else if ( vbo->layout == VBO_LAYOUT_STATIC )
	{
		R_SetAttributeLayoutsStatic( vbo );
	}
	else if ( vbo->layout == VBO_LAYOUT_POSITION )
	{
		R_SetAttributeLayoutsPosition( vbo );
	}
	else
	{
		ri.Error( ERR_DROP, S_COLOR_YELLOW "Unknown attribute layout for vbo: %s\n", vbo->name );
	}
}

// index has to be in range 0-255, weight has to be >= 0 and < 1
static inline unsigned short
boneFactor( int index, float weight ) {
	int scaledWeight = lrintf( weight * 256.0 );
	return (unsigned short)( ( index << 8 ) | MIN( scaledWeight, 255 ) );
}

static void R_CopyVertexData( VBO_t *vbo, byte *outData, vboData_t inData )
{
	uint32_t v;

	if ( vbo->layout == VBO_LAYOUT_VERTEX_ANIMATION )
	{
		struct fmtVertexAnim1 *ptr = ( struct fmtVertexAnim1 * )outData;

		for ( v = 0; v < vbo->framesNum * vbo->vertexesNum; v++ )
		{

			if ( ( vbo->attribBits & ATTR_POSITION ) )
			{
				vec4_t tmp;
				VectorScale( inData.xyz[ v ], 1.0f / 512.0f, tmp);
				tmp[4] = 1.0f; // unused

				floatToSnorm16( tmp, ptr[ v ].position );
			}

			if ( ( vbo->attribBits & ATTR_QTANGENT ) )
			{
				Vector4Copy( inData.qtangent[ v ], ptr[ v ].qtangents );
			}
		}
	}

	for ( v = 0; v < vbo->vertexesNum; v++ )
	{
		if ( vbo->layout == VBO_LAYOUT_SKELETAL ) {
			struct fmtSkeletal *ptr = ( struct fmtSkeletal * )outData;
			if ( ( vbo->attribBits & ATTR_POSITION ) )
			{
				vec4_t tmp;
				VectorCopy( inData.xyz[ v ], tmp);
				tmp[4] = 1.0f; // unused

				floatToSnorm16( tmp, ptr[ v ].position );
			}

			if ( ( vbo->attribBits & ATTR_TEXCOORD ) )
			{
				Vector2Copy( inData.st[ v ], ptr[ v ].texcoord );
			}

			if ( ( vbo->attribBits & ATTR_COLOR ) )
			{
				Vector4Copy( inData.color[ v ], ptr[ v ].colour );
			}

			if ( ( vbo->attribBits & ATTR_QTANGENT ) )
			{
				Vector4Copy( inData.qtangent[ v ], ptr[ v ].qtangents );
			}

			if ( ( vbo->attribBits & ATTR_BONE_FACTORS ) )
			{
				uint32_t j;

				ptr[ v ].boneFactors[ 0 ] = boneFactor( inData.boneIndexes[ v ][ 0 ],
									1.0f - inData.boneWeights[ v ][ 0 ]);
				for ( j = 1; j < 4; j++ ) {
					ptr[ v ].boneFactors[ j ] = boneFactor( inData.boneIndexes[ v ][ j ],
										inData.boneWeights[ v ][ j ] );
				}
			}
		} else if ( vbo->layout == VBO_LAYOUT_STATIC ) {
			shaderVertex_t *ptr = ( shaderVertex_t * )outData;
			if ( ( vbo->attribBits & ATTR_POSITION ) )
			{
				VectorCopy( inData.xyz[ v ], ptr[ v ].xyz );
			}

			if ( ( vbo->attribBits & ATTR_COLOR ) )
			{
				Vector4Copy( inData.color[ v ], ptr[ v ].color );
			}

			if ( ( vbo->attribBits & ATTR_QTANGENT ) )
			{
				Vector4Copy( inData.qtangent[ v ], ptr[ v ].qtangents );
			}

			if ( ( vbo->attribBits & ATTR_TEXCOORD ) )
			{
				Vector4Copy( inData.stpq[ v ], ptr[ v ].texCoords );
			}
		} else if ( vbo->layout == VBO_LAYOUT_POSITION ) {
			vec3_t *ptr = ( vec3_t * )outData;
			if ( ( vbo->attribBits & ATTR_POSITION ) )
			{
				VectorCopy( inData.xyz[ v ], ptr[ v ] );
			}
		} else if ( vbo->layout == VBO_LAYOUT_VERTEX_ANIMATION ) {
			struct fmtVertexAnim2 *ptr = ( struct fmtVertexAnim2 * )( outData + ( vbo->framesNum * vbo->vertexesNum ) * sizeVertexAnim1 );

			if ( ( vbo->attribBits & ATTR_TEXCOORD ) )
			{
				Vector2Copy( inData.st[ v ], ptr[ v ].texcoord );
			}

			if ( ( vbo->attribBits & ATTR_COLOR ) )
			{
				Vector4Copy( inData.color[ v ], ptr[ v ].colour );
			}
		}
	}

}

VBO_t *R_CreateDynamicVBO( const char *name, int numVertexes, uint32_t stateBits, vboLayout_t layout )
{
	VBO_t *vbo;

	if ( !numVertexes )
	{
		return NULL;
	}

	if ( strlen( name ) >= MAX_QPATH )
	{
		ri.Error( ERR_DROP, "R_CreateDynamicVBO: \"%s\" is too long", name );
	}

	// make sure the render thread is stopped
	R_SyncRenderThread();

	vbo = (VBO_t*) ri.Hunk_Alloc( sizeof( *vbo ), h_low );
	memset( vbo, 0, sizeof( *vbo ) );

	Com_AddToGrowList( &tr.vbos, vbo );

	Q_strncpyz( vbo->name, name, sizeof( vbo->name ) );

	vbo->layout = layout;
	vbo->framesNum = 0;
	vbo->vertexesNum = numVertexes;
	vbo->attribBits = stateBits;
	vbo->usage = GL_DYNAMIC_DRAW;

	R_SetVBOAttributeLayouts( vbo, qfalse );

	glGenBuffers( 1, &vbo->vertexesVBO );

	R_BindVBO( vbo );
	glBufferData( GL_ARRAY_BUFFER, vbo->vertexesSize, NULL, vbo->usage );
	R_BindNullVBO();

	GL_CheckErrors();

	return vbo;
}

/*
============
R_CreateVBO
============
*/
VBO_t *R_CreateStaticVBO( const char *name, vboData_t data, vboLayout_t layout )
{
	VBO_t *vbo;
	byte *outData;

	if ( strlen( name ) >= MAX_QPATH )
	{
		ri.Error( ERR_DROP, "R_CreateVBO: \"%s\" is too long", name );
	}

	// make sure the render thread is stopped
	R_SyncRenderThread();

	vbo = (VBO_t*) ri.Hunk_Alloc( sizeof( *vbo ), h_low );
	memset( vbo, 0, sizeof( *vbo ) );

	Com_AddToGrowList( &tr.vbos, vbo );

	Q_strncpyz( vbo->name, name, sizeof( vbo->name ) );

	vbo->layout = layout;
	vbo->vertexesNum = data.numVerts;
	vbo->framesNum = data.numFrames;
	vbo->attribBits = R_DeriveAttrBits( data );
	vbo->usage = GL_STATIC_DRAW;

	R_SetVBOAttributeLayouts( vbo, data.noLightCoords );

	glGenBuffers( 1, &vbo->vertexesVBO );

	R_BindVBO( vbo );
	glBufferData( GL_ARRAY_BUFFER, vbo->vertexesSize, NULL, vbo->usage );
	outData = (byte *)glMapBuffer( GL_ARRAY_BUFFER, GL_WRITE_ONLY );

	R_CopyVertexData( vbo, outData, data );

	glUnmapBuffer( GL_ARRAY_BUFFER );
	R_BindNullVBO();

	GL_CheckErrors();

	return vbo;
}

static vboData_t R_CreateVBOData( const VBO_t *vbo, const srfVert_t *verts, qboolean noLightCoords )
{
	uint32_t v;
	vboData_t data;
	memset( &data, 0, sizeof( data ) );
	data.numVerts = vbo->vertexesNum;
	data.numFrames = vbo->framesNum;
	data.noLightCoords = noLightCoords;

	for ( v = 0; v < vbo->vertexesNum; v++ )
	{
		const srfVert_t *vert = verts + v;
		if ( ( vbo->attribBits & ATTR_POSITION ) )
		{
			if ( !data.xyz )
			{
				data.xyz = ( vec3_t * ) ri.Hunk_AllocateTempMemory( sizeof( *data.xyz ) * data.numVerts );
			}
			VectorCopy( vert->xyz, data.xyz[ v ] );
		}

		if ( ( vbo->attribBits & ATTR_TEXCOORD ) && noLightCoords )
		{
			if ( !data.st )
			{
				data.st = ( i16vec2_t * ) ri.Hunk_AllocateTempMemory( sizeof( i16vec2_t ) * data.numVerts );
			}
			data.st[ v ][ 0 ] = floatToHalf( vert->st[ 0 ] );
			data.st[ v ][ 1 ] = floatToHalf( vert->st[ 1 ] );
		}

		if ( ( vbo->attribBits & ATTR_TEXCOORD ) && !noLightCoords )
		{
			if ( !data.stpq )
			{
				data.stpq = ( i16vec4_t * ) ri.Hunk_AllocateTempMemory( sizeof( i16vec4_t ) * data.numVerts );
			}
			data.stpq[ v ][ 0 ] = floatToHalf( vert->st[ 0 ] );
			data.stpq[ v ][ 1 ] = floatToHalf( vert->st[ 1 ] );
			data.stpq[ v ][ 2 ] = floatToHalf( vert->lightmap[ 0 ] );
			data.stpq[ v ][ 3 ] = floatToHalf( vert->lightmap[ 1 ] );
		}

		if ( ( vbo->attribBits & ATTR_QTANGENT ) )
		{
			if ( !data.qtangent )
			{
				data.qtangent = ( i16vec4_t * ) ri.Hunk_AllocateTempMemory( sizeof( *data.qtangent ) * data.numVerts );
			}
			Vector4Copy( vert->qtangent, data.qtangent[ v ] );
		}

		if ( ( vbo->attribBits & ATTR_COLOR ) )
		{
			if ( !data.color )
			{
				data.color = ( u8vec4_t * ) ri.Hunk_AllocateTempMemory( sizeof( *data.color ) * data.numVerts );
			}
			Vector4Copy( vert->lightColor, data.color[ v ] );
		}
	}

	return data;
}

static void R_FreeVBOData( vboData_t data )
{
	if ( data.color )
	{
		ri.Hunk_FreeTempMemory( data.color );
	}

	if ( data.qtangent )
	{
		ri.Hunk_FreeTempMemory( data.qtangent );
	}

	if ( data.st )
	{
		ri.Hunk_FreeTempMemory( data.st );
	}

	if ( data.xyz )
	{
		ri.Hunk_FreeTempMemory( data.xyz );
	}
}

/*
============
R_CreateVBO2
============
*/
VBO_t *R_CreateStaticVBO2( const char *name, int numVertexes, srfVert_t *verts, unsigned int stateBits )
{
	VBO_t  *vbo;

	byte   *data;
	vboData_t vboData;

	if ( !numVertexes )
	{
		return NULL;
	}

	if ( strlen( name ) >= MAX_QPATH )
	{
		ri.Error( ERR_DROP, "R_CreateVBO2: \"%s\" is too long", name );
	}

	// make sure the render thread is stopped
	R_SyncRenderThread();

	vbo = ( VBO_t * ) ri.Hunk_Alloc( sizeof( *vbo ), h_low );
	memset( vbo, 0, sizeof( *vbo ) );

	Com_AddToGrowList( &tr.vbos, vbo );

	Q_strncpyz( vbo->name, name, sizeof( vbo->name ) );

	vbo->layout = VBO_LAYOUT_STATIC;
	vbo->framesNum = 0;
	vbo->vertexesNum = numVertexes;
	vbo->attribBits = stateBits;
	vbo->usage = GL_STATIC_DRAW;

	vboData = R_CreateVBOData( vbo, verts, qfalse );

	R_SetVBOAttributeLayouts( vbo, qfalse );
	
	data = ( byte * ) ri.Hunk_AllocateTempMemory( vbo->vertexesSize );

	R_CopyVertexData( vbo, data, vboData );

	glGenBuffers( 1, &vbo->vertexesVBO );

	R_BindVBO( vbo );
	glBufferData( GL_ARRAY_BUFFER, vbo->vertexesSize, data, vbo->usage );
	R_BindNullVBO();

	GL_CheckErrors();

	ri.Hunk_FreeTempMemory( data );
	R_FreeVBOData( vboData );

	return vbo;
}

/*
============
R_CreateIBO
============
*/
IBO_t *R_CreateDynamicIBO( const char *name, int numIndexes )
{
	IBO_t *ibo;

	if ( strlen( name ) >= MAX_QPATH )
	{
		ri.Error( ERR_DROP, "R_CreateIBO: \"%s\" is too long", name );
	}

	// make sure the render thread is stopped
	R_SyncRenderThread();

	ibo = (IBO_t*) ri.Hunk_Alloc( sizeof( *ibo ), h_low );
	Com_AddToGrowList( &tr.ibos, ibo );

	Q_strncpyz( ibo->name, name, sizeof( ibo->name ) );

	ibo->indexesSize = numIndexes * sizeof( glIndex_t );
	ibo->indexesNum = numIndexes;

	glGenBuffers( 1, &ibo->indexesVBO );

	R_BindIBO( ibo );
	glBufferData( GL_ELEMENT_ARRAY_BUFFER, ibo->indexesSize, NULL, GL_DYNAMIC_DRAW );
	R_BindNullIBO();

	GL_CheckErrors();

	return ibo;
}

/*
============
R_CreateIBO2
============
*/
IBO_t *R_CreateStaticIBO( const char *name, glIndex_t *indexes, int numIndexes )
{
	IBO_t         *ibo;

	if ( !numIndexes )
	{
		return NULL;
	}

	if ( strlen( name ) >= MAX_QPATH )
	{
		ri.Error( ERR_DROP, "R_CreateIBO: \"%s\" is too long", name );
	}

	// make sure the render thread is stopped
	R_SyncRenderThread();

	ibo = ( IBO_t * ) ri.Hunk_Alloc( sizeof( *ibo ), h_low );
	Com_AddToGrowList( &tr.ibos, ibo );

	Q_strncpyz( ibo->name, name, sizeof( ibo->name ) );

	ibo->indexesSize = numIndexes * sizeof( glIndex_t );
	ibo->indexesNum = numIndexes;

	glGenBuffers( 1, &ibo->indexesVBO );

	R_BindIBO( ibo );
	glBufferData( GL_ELEMENT_ARRAY_BUFFER, ibo->indexesSize, indexes, GL_STATIC_DRAW );
	R_BindNullIBO();

	GL_CheckErrors();

	return ibo;
}

IBO_t *R_CreateStaticIBO2( const char *name, int numTriangles, srfTriangle_t *triangles )
{
	IBO_t         *ibo;
	int           i, j;

	glIndex_t     *indexes;

	srfTriangle_t *tri;

	if ( !numTriangles )
	{
		return NULL;
	}

	if ( strlen( name ) >= MAX_QPATH )
	{
		ri.Error( ERR_DROP, "R_CreateIBO2: \"%s\" is too long", name );
	}

	// make sure the render thread is stopped
	R_SyncRenderThread();

	indexes = ( glIndex_t * ) ri.Hunk_AllocateTempMemory( numTriangles * 3 * sizeof( glIndex_t ) );

	for ( i = 0, tri = triangles; i < numTriangles; i++, tri++ )
	{
		for ( j = 0; j < 3; j++ )
		{
			indexes[ i * 3 + j ] = tri->indexes[ j ];
		}
	}

	ibo = R_CreateStaticIBO( name, indexes, numTriangles * 3 );

	ri.Hunk_FreeTempMemory( indexes );

	return ibo;
}

/*
============
R_BindVBO
============
*/
void R_BindVBO( VBO_t *vbo )
{
	if ( !vbo )
	{
		ri.Error( ERR_DROP, "R_BindNullVBO: NULL vbo" );
	}

	if ( r_logFile->integer )
	{
		// don't just call LogComment, or we will get a call to va() every frame!
		GLimp_LogComment( va( "--- R_BindVBO( %s ) ---\n", vbo->name ) );
	}

	if ( glState.currentVBO != vbo )
	{
		glState.currentVBO = vbo;
		glState.vertexAttribPointersSet = 0;

		glState.vertexAttribsInterpolation = -1;
		glState.vertexAttribsOldFrame = 0;
		glState.vertexAttribsNewFrame = 0;

		glBindBuffer( GL_ARRAY_BUFFER, vbo->vertexesVBO );

		backEnd.pc.c_vboVertexBuffers++;
	}
}

/*
============
R_BindNullVBO
============
*/
void R_BindNullVBO( void )
{
	GLimp_LogComment( "--- R_BindNullVBO ---\n" );

	if ( glState.currentVBO )
	{
		glBindBuffer( GL_ARRAY_BUFFER, 0 );
		glState.currentVBO = NULL;
	}

	GL_CheckErrors();
}

/*
============
R_BindIBO
============
*/
void R_BindIBO( IBO_t *ibo )
{
	if ( !ibo )
	{
		ri.Error( ERR_DROP, "R_BindIBO: NULL ibo" );
	}

	if ( r_logFile->integer )
	{
		// don't just call LogComment, or we will get a call to va() every frame!
		GLimp_LogComment( va( "--- R_BindIBO( %s ) ---\n", ibo->name ) );
	}

	if ( glState.currentIBO != ibo )
	{
		glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, ibo->indexesVBO );

		glState.currentIBO = ibo;

		backEnd.pc.c_vboIndexBuffers++;
	}
}

/*
============
R_BindNullIBO
============
*/
void R_BindNullIBO( void )
{
	GLimp_LogComment( "--- R_BindNullIBO ---\n" );

	if ( glState.currentIBO )
	{
		glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, 0 );
		glState.currentIBO = NULL;
		glState.vertexAttribPointersSet = 0;
	}
}

static void R_InitUnitCubeVBO( void )
{
	vec3_t        mins = { -1, -1, -1 };
	vec3_t        maxs = { 1,  1,  1 };
	vboData_t     data;
	int           i;

	R_SyncRenderThread();

	tess.multiDrawPrimitives = 0;
	tess.numIndexes = 0;
	tess.numVertexes = 0;

	Tess_AddCube( vec3_origin, mins, maxs, colorWhite );

	memset( &data, 0, sizeof( data ) );
	data.xyz = ( vec3_t * ) ri.Hunk_AllocateTempMemory( tess.numVertexes * sizeof( *data.xyz ) );

	data.numVerts = tess.numVertexes;

	for ( i = 0; i < tess.numVertexes; i++ )
	{
		VectorCopy( tess.verts[ i ].xyz, data.xyz[ i ] );
	}

	tr.unitCubeVBO = R_CreateStaticVBO( "unitCube_VBO", data, VBO_LAYOUT_POSITION );
	tr.unitCubeIBO = R_CreateStaticIBO( "unitCube_IBO", tess.indexes, tess.numIndexes );

	ri.Hunk_FreeTempMemory( data.xyz );

	tess.multiDrawPrimitives = 0;
	tess.numIndexes = 0;
	tess.numVertexes = 0;
}

/*
============
R_InitVBOs
============
*/
void R_InitVBOs( void )
{
	uint32_t attribs = ATTR_POSITION | ATTR_TEXCOORD | ATTR_QTANGENT | ATTR_COLOR;

	ri.Printf( PRINT_DEVELOPER, "------- R_InitVBOs -------\n" );

	Com_InitGrowList( &tr.vbos, 100 );
	Com_InitGrowList( &tr.ibos, 100 );

	tess.vbo = R_CreateDynamicVBO( "tessVertexArray_VBO", SHADER_MAX_VERTEXES, attribs, VBO_LAYOUT_STATIC );

	tess.ibo = R_CreateDynamicIBO( "tessVertexArray_IBO", SHADER_MAX_INDEXES );

	R_InitUnitCubeVBO();

	// allocate a PBO for color grade map transfers
	glGenBuffers( 1, &tr.colorGradePBO );
	glBindBuffer( GL_PIXEL_PACK_BUFFER, tr.colorGradePBO );
	glBufferData( GL_PIXEL_PACK_BUFFER,
		      REF_COLORGRADEMAP_STORE_SIZE * sizeof(u8vec4_t),
		      NULL, GL_STREAM_COPY );
	glBindBuffer( GL_PIXEL_PACK_BUFFER, 0 );

	GL_CheckErrors();
}

/*
============
R_ShutdownVBOs
============
*/
void R_ShutdownVBOs( void )
{
	int   i;
	VBO_t *vbo;
	IBO_t *ibo;

	ri.Printf( PRINT_DEVELOPER, "------- R_ShutdownVBOs -------\n" );

	R_BindNullVBO();
	R_BindNullIBO();

	glDeleteBuffers( 1, &tr.colorGradePBO );

	for ( i = 0; i < tr.vbos.currentElements; i++ )
	{
		vbo = ( VBO_t * ) Com_GrowListElement( &tr.vbos, i );

		if ( vbo->vertexesVBO )
		{
			glDeleteBuffers( 1, &vbo->vertexesVBO );
		}
	}

	for ( i = 0; i < tr.ibos.currentElements; i++ )
	{
		ibo = ( IBO_t * ) Com_GrowListElement( &tr.ibos, i );

		if ( ibo->indexesVBO )
		{
			glDeleteBuffers( 1, &ibo->indexesVBO );
		}
	}

	Com_DestroyGrowList( &tr.vbos );
	Com_DestroyGrowList( &tr.ibos );
}

/*
==============
Tess_UpdateVBOs

Tr3B: update the default VBO to replace the client side vertex arrays
==============
*/
void Tess_UpdateVBOs( uint32_t attribBits )
{
	if ( r_logFile->integer )
	{
		GLimp_LogComment( va( "--- Tess_UpdateVBOs( attribBits = %i ) ---\n", attribBits ) );
	}

	GL_CheckErrors();

	// update the default VBO
	if ( tess.numVertexes > 0 && tess.numVertexes <= SHADER_MAX_VERTEXES )
	{
		R_BindVBO( tess.vbo );

		GL_CheckErrors();

		assert( ( attribBits & ATTR_BITS ) != 0 );

		GL_VertexAttribsState( attribBits );

		if ( r_logFile->integer )
		{
			GLimp_LogComment( va( "glBufferSubData( vbo = '%s', numVertexes = %i )\n", tess.vbo->name, tess.numVertexes ) );
		}

		glBufferSubData( GL_ARRAY_BUFFER, 0, tess.numVertexes * sizeof( shaderVertex_t ), tess.verts );
	}

	GL_CheckErrors();

	// update the default IBO
	if ( tess.numIndexes > 0 && tess.numIndexes <= SHADER_MAX_INDEXES )
	{
		R_BindIBO( tess.ibo );

		glBufferSubData( GL_ELEMENT_ARRAY_BUFFER, 0, tess.numIndexes * sizeof( glIndex_t ), tess.indexes );
	}

	GL_CheckErrors();
}

/*
============
R_VBOList_f
============
*/
void R_VBOList_f( void )
{
	int   i;
	VBO_t *vbo;
	IBO_t *ibo;
	int   vertexesSize = 0;
	int   indexesSize = 0;

	ri.Printf( PRINT_ALL, " size          name\n" );
	ri.Printf( PRINT_ALL, "----------------------------------------------------------\n" );

	for ( i = 0; i < tr.vbos.currentElements; i++ )
	{
		vbo = ( VBO_t * ) Com_GrowListElement( &tr.vbos, i );

		ri.Printf( PRINT_ALL, "%d.%02d MB %s\n", vbo->vertexesSize / ( 1024 * 1024 ),
		           ( vbo->vertexesSize % ( 1024 * 1024 ) ) * 100 / ( 1024 * 1024 ), vbo->name );

		vertexesSize += vbo->vertexesSize;
	}

	for ( i = 0; i < tr.ibos.currentElements; i++ )
	{
		ibo = ( IBO_t * ) Com_GrowListElement( &tr.ibos, i );

		ri.Printf( PRINT_ALL, "%d.%02d MB %s\n", ibo->indexesSize / ( 1024 * 1024 ),
		           ( ibo->indexesSize % ( 1024 * 1024 ) ) * 100 / ( 1024 * 1024 ), ibo->name );

		indexesSize += ibo->indexesSize;
	}

	ri.Printf( PRINT_ALL, " %i total VBOs\n", tr.vbos.currentElements );
	ri.Printf( PRINT_ALL, " %d.%02d MB total vertices memory\n", vertexesSize / ( 1024 * 1024 ),
	           ( vertexesSize % ( 1024 * 1024 ) ) * 100 / ( 1024 * 1024 ) );

	ri.Printf( PRINT_ALL, " %i total IBOs\n", tr.ibos.currentElements );
	ri.Printf( PRINT_ALL, " %d.%02d MB total triangle indices memory\n", indexesSize / ( 1024 * 1024 ),
	           ( indexesSize % ( 1024 * 1024 ) ) * 100 / ( 1024 * 1024 ) );
}
